/*
 * BaseStation.hpp
 *
 *  Created on: Aug 26, 2011
 *      Author: ra23501
 */

#ifndef BASESTATION_HPP_
#define BASESTATION_HPP_

#include <ros/ros.h>
#include <ros/console.h>
#include <boost/thread/thread.hpp>
#include <geometry_msgs/Vector3.h>
#include <geometry_msgs/Quaternion.h>
#include <wpi_msgs/LandRequest.h>
#include <wpi_msgs/LandConfirm.h>
#include <wpi_msgs/RobotState.h>

#include "BatteryPack.hpp"
#include "MaestroController.hpp"

// define number of batteries in system
#ifndef NUM_BATTERIES
#define NUM_BATTERIES 8
#endif

// define pi
#ifndef PI
#define PI 3.14159265358979323846
#endif

// define epsilon
#define EPSILON 1E-6

// macros
#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

#ifndef MINMAX
#define MINMAX(A,B,C) MIN((B),MAX((A),(C)))
#endif

#ifndef NORMALIZE
#define NORMALIZE(Z) atan2(sin(Z),cos(Z))
#endif

// enums
enum BaseState
{
  IDLE, LANDING, CENTERING, REMOVING_BATTERY, REPLACING_BATTERY, LAUNCHING
};

using namespace std;

class BaseStation
{
public:
  BaseStation();
  ~BaseStation();

private:
  // callbacks
  void callbackStateUpdate(const boost::shared_ptr<wpi_msgs::RobotState const> &robot_state_msg);
  void callbackLandRequestMsg(const boost::shared_ptr<wpi_msgs::LandRequest const> &land_request_msg);
  void callbackLandConfirmMsg(const boost::shared_ptr<wpi_msgs::LandConfirm const> &land_confirm_msg);

  // init functions
  void initParams(void);
  void initSubscribers(void);
  void initBatteries(void);

  // robot states
  void idleState(void);
  void landingState(void);
  void centeringState(void);
  void removingBatteryState(void);
  void replacingBatteryState(void);
  void launchingState(void);

  // base station thread
  void startBaseThread(void);
  void baseThread(void);
  void stopBaseThread(void);

  // listener thread
  void startListenThread(void);
  void listenThread(void);
  void stopListenThread(void);

  // thread helper functions
  void findNodes(void);
  void addNode(string name);
  float findRemainingTime(string name);

  // ros objects
  ros::NodeHandle nh_;
  ros::NodeHandle private_nh_;

  // ros param
  string uid_;
  double base_thread_rate_;
  double max_battery_voltage_;
  double return_battery_voltage_;
  double min_battery_voltage_;

  // ros publishers
  std::vector<ros::Publisher> land_request_pub_;
  std::vector<ros::Publisher> land_confirm_pub_;

  // ros subscribers
  ros::Subscriber robot_state_sub_;
  ros::Subscriber land_request_sub_;
  ros::Subscriber land_confirm_sub_;

  // threads
  boost::shared_ptr<boost::thread> robot_state_sub_thread_;
  boost::shared_ptr<boost::thread> base_thread_;
  boost::shared_ptr<boost::thread> listen_thread_;
  boost::mutex mutex_;

  // general containers
  wpi_msgs::RobotState robot_state_;
  std::vector<std::string> all_UAVs_;
  std::vector<std::string> waiting_UAVs_;
  BaseState base_state_;
  std::string current_UAV_;
  MaestroController controller;
  BatteryPack* allBatteries[NUM_BATTERIES];
  std::vector<BatteryPack*> readyBatteries;
  char openStation;

  // volatile bools
  volatile bool received_robot_update_;
  volatile bool received_land_request_;
  volatile bool received_positive_land_confirm_;
  volatile bool received_negative_land_confirm_;
  volatile bool UAV_landing_;
};

#endif /* BASESTATION_HPP_ */
