/*
 * MaestroController.hpp
 *
 *  Created on: Sep 14, 2011
 *      Author: ra23501
 */

/*
 * Portions of this code including this header file and source files prefixed
 * with MaestroController_ contain code from at least one of libusb-1.0 and
 * the Pololu USB SDK. Libusb-1.0 is licensed under terms of the LGPL found at
 * http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html and the Pololu SDK
 * is licensed under their license, found in the file PololuLicense.txt and at
 * http://www.pololu.com/docs/0J41.
 */

#ifndef MAESTRO_CONTROLLER_HPP
#define MAESTRO_CONTROLLER_HPP

#include <libusb.h>
#include <ros/ros.h>

// defines
#define VENDOR_POLOLU 0x1ffb
#define MICRO_MAESTRO_STACK_SIZE  32
#define MICRO_MAESTRO_CALL_STACK_SIZE 10
#define MINI_MAESTRO_STACK_SIZE 126
#define MINI_MAESTRO_CALL_STACK_SIZE 126
// for Maestro Mini 24
#define SERVO_COUNT 24
// stall is checked at 2Hz
#define STALL_OUT 2
#define NUM_BATTERIES 8
// logic high/low, also directions
#define STEPPER_CCW 2000*4
#define STEPPER_CW 1000*4
#define TRUE 2000
#define FALSE 1000
#define ON 2000*4
#define OFF 1000*4
#define ON_INP 1023
#define OFF_INP 0
// speeds for cart rack servo
#define CART_RIGHT 1600*4
#define CART_LEFT 1400*4
#define STOP 1500*4
// LED color cutoffs
#define GREEN 770
#define ORANGE 255

// macros
#ifndef VTOSIG
#define VTOSIG(A) (A * 1023 / 5)
#endif

#ifndef SIGTOV
#define SIGTOV(A) (A * 5 / 1023)
#endif

#ifndef ISGREEN
#define ISGREEN(A) (A >= GREEN)
#endif

#ifndef ISORANGE
#define ISORANGE(A) (A >= ORANGE && A < GREEN)
#endif

#ifndef ISRED
#define ISRED(A) (A < ORANGE)
#endif

// enums
// product codes to validate USB devices
enum MaestroProduct
{
  MAESTRO_PROD_MICRO = 0x0089, MAESTRO_PROD_MINI_12 = 0x008A, MAESTRO_PROD_MINI_18 = 0x008B,
  MAESTRO_PROD_MINI_24 = 0x008C
};

// serial numbers for devices to discern them from each other
enum MaestroSerial
{
  MICRO_SERIAL = 7021, PRIMARY_SERIAL = 17565, SECONDARY_SERIAL = 17571
};

/*enum ScissorPotValues {
 SCISSOR_REST_POT = ,
 SCISSOR_UAV_POT = ,
 SCISSOR_CHARGE_POT = ,
 SCISSOR_DOWN_POT = 142
 };*/

// These are the values to put in to bRequest when making a setup packet
// for a control transfer to the Maestro.  See the comments and code in Usc.cs
// for more information about what these requests do and the format of the
// setup packet.
enum Request
{
  REQUEST_GET_PARAMETER = 0x81, REQUEST_SET_PARAMETER = 0x82, REQUEST_GET_VARIABLES = 0x83,
  REQUEST_SET_SERVO_VARIABLE = 0x84, REQUEST_SET_TARGET = 0x85,
  REQUEST_CLEAR_ERRORS = 0x86,
  REQUEST_GET_SERVO_SETTINGS = 0x87,

  // GET STACK and GET CALL STACK are only used on the Mini Maestro.
  REQUEST_GET_STACK = 0x88, REQUEST_GET_CALL_STACK = 0x89, REQUEST_SET_PWM = 0x8A,

  REQUEST_REINITIALIZE = 0x90, REQUEST_ERASE_SCRIPT = 0xA0,
  REQUEST_WRITE_SCRIPT = 0xA1,
  REQUEST_SET_SCRIPT_DONE = 0xA2, // value.low.b is 0 for go, 1 for stop, 2 for single-step
  REQUEST_RESTART_SCRIPT_AT_SUBROUTINE = 0xA3, REQUEST_RESTART_SCRIPT_AT_SUBROUTINE_WITH_PARAMETER = 0xA4,
  REQUEST_RESTART_SCRIPT = 0xA5, REQUEST_START_BOOTLOADER = 0xFF
};

// channels for the Maestro Micro controller
enum MicroChannels
{
  MICRO_TEST_SERVO = 0, MICRO_SECOND_SERVO = 1, MICRO_LIMIT = 4, MICRO_PHOTO = 5
};

// channels for the Maestro Mini 24 primary (servo) controller
enum PrimaryMiniChannels
{
  ARM_SERVO_0 = 0, ARM_SERVO_1 = 1, CART_0_SERVO = 2, CART_0_MOTOR = 3, CART_1_SERVO = 4, SCISSOR_POT = 5,
  CART_END_0 = 6, CART_END_1 = 7, CART_MID = 8, STEPPER_DIRECTION = 9, STEPPER_STEPS = 10, STEPPER_HOME = 15,
  BATTERY_STATION_0 = 11, BATTERY_STATION_1 = 12, BATTERY_STATION_2 = 13, BATTERY_STATION_3 = 14,
  BATTERY_STATION_4 = 15, BATTERY_STATION_5 = 16, BATTERY_STATION_6 = 17, BATTERY_STATION_7 = 18,
  ARM_SERVO_0_CURRENT = 19, ARM_SERVO_1_CURRENT = 20, CART_0_SERVO_CURRENT = 21, CART_1_SERVO_CURRENT = 22
};

// channels for the Maestro Mini 24 secondary (battery) controller
enum SecondaryMiniChannels
{
  BATTERY_START_RELAY_0 = 0, BATTERY_START_RELAY_1 = 1, BATTERY_START_RELAY_2 = 2, BATTERY_START_RELAY_3 = 3,
  BATTERY_START_RELAY_4 = 4, BATTERY_START_RELAY_5 = 5, BATTERY_START_RELAY_6 = 6, BATTERY_START_RELAY_7 = 7,
  BATTERY_LED_0 = 8, BATTERY_LED_1 = 9, BATTERY_LED_2 = 10, BATTERY_LED_3 = 11, BATTERY_LED_4 = 12, BATTERY_LED_5 = 13,
  BATTERY_LED_6 = 14, BATTERY_LED_7 = 15, BATTERY_PRESENT_0 = 16, BATTERY_PRESENT_1 = 17, BATTERY_PRESENT_2 = 18,
  BATTERY_PRESENT_3 = 19, BATTERY_PRESENT_4 = 20, BATTERY_PRESENT_5 = 21, BATTERY_PRESENT_6 = 22,
  BATTERY_PRESENT_7 = 23
};

// hard limits for the servos, by name
enum ServoLimits
{
  GWS_MAX = 2480 * 4, GWS_MIN = 592 * 4
};

// some commonly-used servo values
enum ServoValues
{
  SCISSOR_REST = 1295 * 4,
  SCISSOR_UAV = 1225 * 4,
  SCISSOR_CHARGE = 1605 * 4,
  SCISSOR_DOWN = 1520 * 4,
  SCISSOR_MAX_UP = 800 * 4, SCISSOR_MAX_DOWN = 2100 * 4

/*
 GWS_HOME = 2480 * 4, SCISSOR_REST = 1545 * 4, // 87.5
 SCISSOR_UAV = 1500 * 4, // 85.0
 SCISSOR_CHARGE = 1850 * 4, // 104.5
 SCISSOR_DOWN = 1975 * 4, // 108.5
 SCISSOR_MAX_UP = 1200 * 4, SCISSOR_MAX_DOWN = 2300 * 4
 */

};

// Represents the current status of a channel.
struct ServoStatus
{
  // The position in units of quarter-microseconds.
  unsigned short position;

  // The target position in units of quarter-microseconds.
  unsigned short target;

  // The speed limit.
  unsigned short speed;

  // The acceleration limit.
  unsigned char acceleration;
};
typedef struct ServoStatus ServoStatus;

// Represents the variables that can be read from
// a Micro Maestro or Mini Maestro using REQUEST_GET_VARIABLES,
// excluding channel information, the stack, and the call stack.
struct MaestroVariables
{
  // The number of values on the data stack (0-32).  A value of 0 means the stack is empty.
  unsigned char stackPointer;

  // The number of return locations on the call stack (0-10).  A value of 0 means the stack is empty.
  unsigned char callStackPointer;

  // The error register.  Each bit stands for a different error (see uscError).
  // If the bit is one, then it means that error occurred some time since the last
  // GET_ERRORS serial command or CLEAR_ERRORS USB command.
  unsigned short errors;

  // The address (in bytes) of the next bytecode instruction that will be executed.
  unsigned short programCounter;

  // 0 = script is running.
  // 1 = script is done.
  // 2 = script will be done as soon as it executes one more instruction
  //     (used to implement step-through debugging features)
  unsigned char scriptDone;

  // The performance flag register.  Each bit represents a different flag.
  // If it is 1, then it means that the flag occurred some time since the last
  // getVariables request.  This register is always 0 for the Micro Maestro
  // because performance flags only apply to the Mini Maestros.
  unsigned char performanceFlags;
};
typedef struct MaestroVariables MaestroVariables;

// Represents the non-channel-specific variables that can be read from
// a Micro Maestro using REQUEST_GET_VARIABLES.
struct MicroMaestroVariables
{
  // The number of values on the data stack (0-32).  A value of 0 means the stack is empty.
  unsigned char stackPointer;

  // The number of return locations on the call stack (0-10).  A value of 0 means the stack is empty.
  unsigned char callStackPointer;

  // The error register.  Each bit stands for a different error (see uscError).
  // If the bit is one, then it means that error occurred some time since the last
  // GET_ERRORS serial command or CLEAR_ERRORS USB command.
  unsigned short errors;

  // The address (in bytes) of the next bytecode instruction that will be executed.
  unsigned short programCounter;

  // Meaningless bytes to protect the program from stack underflows.
  unsigned short buffer[3];

  // The data stack used by the script.  The values in locations 0 through stackPointer-1
  // are on the stack.
  unsigned short stack[32];

  // The call stack used by the script.  The addresses in locations 0 through
  // callStackPointer-1 are on the call stack.  The next return will make the
  // program counter go to callStack[callStackPointer-1].
  unsigned short callStack[10];

  // 0 = script is running.
  // 1 = script is done.
  // 2 = script will be done as soon as it executes one more instruction
  //     (used to implement step-through debugging features)
  unsigned char scriptDone;

  // Meaningless byte to protect the program from call stack overflows.
  unsigned char buffer2;

  // Container for channel information. The struct is 1 byte too large (should be 7, not 8)
  ServoStatus ServoSettings[6];
};
typedef struct MicroMaestroVariables MicroMaestroVariables;

// Represents the variables that can be read from a Mini Maestro using REQUEST_GET_VARIABLES.
struct MiniMaestroVariables
{
  // The number of values on the data stack (0-32).  A value of 0 means the stack is empty.
  unsigned char stackPointer;

  // The number of return locations on the call stack (0-10).  A value of 0 means the stack is empty.
  unsigned char callStackPointer;

  // The error register.  Each bit stands for a different error (see uscError).
  // If the bit is one, then it means that error occurred some time since the last
  // GET_ERRORS serial command or CLEAR_ERRORS USB command.
  unsigned short errors;

  // The address (in bytes) of the next bytecode instruction that will be executed.
  unsigned short programCounter;

  // 0 = script is running.
  // 1 = script is done.
  // 2 = script will be done as soon as it executes one more instruction
  //     (used to implement step-through debugging features)
  unsigned char scriptDone;

  // The performance flag register.  Each bit represents a different error.
  // If it is 1, then it means that the error occurred some time since the last
  // getVariables request.
  unsigned char performanceFlags;
};
typedef struct MiniMaestroVariables MiniMaestroVariables;

using namespace std;

class MaestroController
{
public:
  MaestroController(void);
  ~MaestroController(void);

  // system actions
  void initializeSystem(void);
  void centerUAV(void);
  void removeBatteryUAV(void);
  void chargeSpentBattery(char openStation);
  void getChargedBattery(char chargedStation);
  void replaceBatteryUAV(void);
  void retractArms(void);
  void disableAll(void);

  // battery functions that the station must call
  bool batteryPresent(char station);
  char doneCharging(void);

private:
  // stepper functions
  void homeStepper(void);
  void moveToStation(char target);
  char getStepperLocation(void);
  void moveCCW(char stations);
  void moveCW(char stations);
  void wiggle(ros::Rate timer);
  void shiftCW(void);
  void shiftCCW(void);

  // cart functions
  void centerCart(void);
  void moveCartLeft(void);
  void moveCartRight(void);
  int moveScissorLift(short height, bool up);

  // servo functions
  int moveServos(short* serial, short* channel, short* value, char numServos);
  void setTarget(short serial, short value, short channel);
  void setSpeed(short serial, short value, short channel);
  void setAcceleration(short serial, short value, short channel);
  void disableServo(short serial, short channel);

  // battery functions
  int startCharger(char station);

  // low-level functions
  void initDevices(void);
  void closeDevices(void);
  void openDevices(bool print);
  int controlTransfer(short serial, unsigned char RequestType, unsigned char Request, short Value, short Index);
  int controlTransfer(short serial, unsigned char RequestType, unsigned char Request, short Value, short Index,
                      void * data, short length);

  // input-/status-related functions
  void getChannelValues(short* serial, short* channel, short* positions, short numServos);
  void displayStatus(short serial);
  void getVariables(short serial, MaestroVariables variables, short* stack, short* callStack, ServoStatus* servos);
  void getVariables(short serial, MaestroVariables variables);
  void getVariables(short serial, ServoStatus* servos);
  void getVariables(short serial, short* stack, bool isStack);
  void getVariablesMicroMaestro(short serial, MaestroVariables variables, short* stack, short* callStack,
                                ServoStatus* servos);
  void getVariablesMiniMaestro(short serial, MaestroVariables variables);
  void getVariablesMiniMaestro(short serial, ServoStatus* servos);
  void getVariablesMiniMaestro(short serial, short* stack, bool isStack);

  // containers
  libusb_device* primary_dev, *secondary_dev;
  libusb_device_handle* primary_hand, *secondary_hand;
  char stepperAt, stepsPerStation;
};
#endif // MAESTRO_CONTROLLER_HPP
