/*
 * BaseStation_thread.cpp
 *
 *  Created on: Aug 31, 2011
 *      Author: ra23501
 */

// includes
#include "BaseStation.hpp"

// start base thread
void BaseStation::startBaseThread(void)
{
  base_thread_ = boost::shared_ptr < boost::thread > (new boost::thread(boost::bind(&BaseStation::baseThread, this)));
}

// base thread
void BaseStation::baseThread(void)
{
  // set thread rate
  ros::Rate thread_rate(base_thread_rate_);

  // while node active
  while (ros::ok())
  {
    // base state machine
    mutex_.lock();
    switch (base_state_)
    {
      case (IDLE):
        idleState();
        break;
      case (LANDING):
        landingState();
        break;
      case (CENTERING):
        centeringState();
        break;
      case (REMOVING_BATTERY):
        removingBatteryState();
        break;
      case (REPLACING_BATTERY):
        replacingBatteryState();
        break;
      case (LAUNCHING):
        launchingState();
        break;
      default:
        break;
    }
    mutex_.unlock();

    // regulate thread rate
    thread_rate.sleep();
  }
}

// stop base thread
void BaseStation::stopBaseThread(void)
{
  base_thread_->join();
}

// start listen thread
void BaseStation::startListenThread(void)
{
  listen_thread_ = boost::shared_ptr < boost::thread
      > (new boost::thread(boost::bind(&BaseStation::listenThread, this)));
}

// listens for requests and status updates from all UAVs in communication
// publishes landing requests/confirmations in applicable cases
void BaseStation::listenThread(void)
{
  // set thread rate
  ros::Rate thread_rate(base_thread_rate_);

  // while node active
  while (ros::ok())
  {
    // base state machine
    mutex_.lock();
    findNodes();
    if (received_land_request_) // one UAV wants to land
    {
      // reset bools
      received_land_request_ = false;

      // publish confirm message
      wpi_msgs::LandConfirm land_confirm_msg;
      land_confirm_msg.header.stamp = ros::Time::now();
      land_confirm_msg.uid = uid_;
      land_confirm_msg.confirm = true;
      for (unsigned int i = 0; i < land_confirm_pub_.size(); i++)
      {
        if (land_confirm_pub_.at(i).getTopic().find(current_UAV_) != string::npos)
        {
          land_confirm_pub_.at(i).publish(land_confirm_msg);
          break;
        }
      }
    }
    else if ((waiting_UAVs_.size() > 0 && !UAV_landing_)) // UAVs are waiting and the base platform is open
    {
      // reset bools
      received_land_request_ = false;
      received_positive_land_confirm_ = false;
      UAV_landing_ = true;

      //publish confirm message
      wpi_msgs::LandConfirm land_confirm_msg;
      land_confirm_msg.header.stamp = ros::Time::now();
      land_confirm_msg.uid = uid_;
      land_confirm_msg.confirm = true;
      for (unsigned int i = 0; i < land_confirm_pub_.size(); i++)
      {
        if (land_confirm_pub_.at(i).getTopic().find(waiting_UAVs_.at(0)) != string::npos)
        {
          land_confirm_pub_.at(i).publish(land_confirm_msg);
          waiting_UAVs_.erase(waiting_UAVs_.begin());
          break;
        }
      }

      // change state
      base_state_ = LANDING;
    }
    else if (received_positive_land_confirm_ && !UAV_landing_) // a UAV agreed to land and the base is open
    {
      // reset bools
      received_positive_land_confirm_ = false;
      UAV_landing_ = true;

      // change state
      base_state_ = LANDING;
    }
    else if (received_robot_update_)
    {
      // reset bools
      received_robot_update_ = false;
      if (robot_state_.battery_voltage <= return_battery_voltage_ && !UAV_landing_) // UAV has low voltage, request it to land
      {
        // publish request message
        wpi_msgs::LandRequest land_request_msg;
        land_request_msg.header.stamp = ros::Time::now();
        land_request_msg.uid = uid_;
        for (unsigned int i = 0; i < land_request_pub_.size(); i++)
        {
          if (land_request_pub_.at(i).getTopic().find(current_UAV_) != std::string::npos)
          {
            land_request_pub_.at(i).publish(land_request_msg);
            UAV_landing_ = true; // UAVs are programmed to always land when requested from the base
            break;
          }
        }
      }
    }
    mutex_.unlock();

    // regulate thread rate
    thread_rate.sleep();
  }
}

// stop listen thread
void BaseStation::stopListenThread(void)
{
  listen_thread_->join();
}
