/*
 * MaestroController_low_level.cpp
 *
 *  Created on: Sep 14, 2011
 *      Author: ra23501
 */

#include "MaestroController.hpp"

// start up USB interfacing
void MaestroController::initDevices(void)
{
  if (libusb_init(NULL) < 0)
    ROS_ERROR("Error initializing");
}

// close any USB interfacing and stop communications
void MaestroController::closeDevices(void)
{
  libusb_exit(NULL);
}

// searches for all available USB devices
// if any are found that carry the vendor code for Pololu, they are added to a list of controller devices
// as one of the known devices. If an unknown device is found, it is labelled as a Micro
// can print out identification information for all UAB devices if boolean print is true
// pre conditions: USB devices have been initialized
void MaestroController::openDevices(bool print)
{
  libusb_device* dev, **devs;
  libusb_device_descriptor desc;
  libusb_device_handle* hand;
  int i = 0;
  bool search = false;

  if (libusb_get_device_list(NULL, &devs) < 0)
    ROS_ERROR("No devices found");

  while ((dev = devs[i++]) != NULL)
  {
    if (libusb_get_device_descriptor(dev, &desc) < 0)
    {
      ROS_ERROR("failed to get device descriptor");
      return;
    }

    if (desc.idVendor == VENDOR_POLOLU)
    {
      libusb_open(dev, &hand);
      // useful for when a device times out
      // resets the device and looks for devices again
      if (libusb_reset_device(hand) == LIBUSB_ERROR_NOT_FOUND)
      {
        libusb_close(hand);
        search = true;
      }
    }
  }

  if (search) // if a device is reset, the devices must be found again
  {
    libusb_free_device_list(devs, 1);
    if (libusb_get_device_list(NULL, &devs) < 0)
      ROS_ERROR("No devices found");
  }

  i = 0; // reset from previous while loop
  while ((dev = devs[i++]) != NULL)
  {
    if (libusb_get_device_descriptor(dev, &desc) < 0)
    {
      ROS_ERROR("failed to get device descriptor");
      return;
    }

    if (desc.idVendor == VENDOR_POLOLU)
    {
      libusb_device_handle* temp_hand;
      libusb_open(dev, &temp_hand);
      int length = 0;
      char buffer[100];
      unsigned char * p = (unsigned char*)buffer;
      // get serial number to use for identification
      length = libusb_get_string_descriptor_ascii(temp_hand, desc.iSerialNumber, p, 100);

      string serial_number = "";
      for (int i = 0; i < length; i++)
      {
        serial_number += (char)buffer[i];
      }
      int ser_num = atoi(serial_number.c_str());

      ROS_ERROR("Serial number is %d", ser_num);
      switch (ser_num)
      {
        case (MICRO_SERIAL):
          primary_dev = dev;
          libusb_open(primary_dev, &primary_hand);
          ROS_ERROR("Micro");
          break;
        case (PRIMARY_SERIAL):
          primary_dev = dev;
          libusb_open(primary_dev, &primary_hand);
          ROS_ERROR("Primary");
          break;
        case (SECONDARY_SERIAL):
          secondary_dev = dev;
          libusb_open(secondary_dev, &secondary_hand);
          ROS_ERROR("Secondary");
          break;
        default:
          primary_dev = dev;
          libusb_open(primary_dev, &primary_hand);
          ROS_ERROR("Micro default");
          break;
      }
    }
    if (print) // print out information on all devices if chosen
    {
      ROS_ERROR(
          "%04x:%04x (bus %d, device %d)\n", desc.idVendor, desc.idProduct, libusb_get_bus_number(dev), libusb_get_device_address(dev));
    }
  }
  libusb_free_device_list(devs, 1); // device list is no longer needed
}

// send a simple command to the controller
// most useful with servo settings and outputs
// pre conditions: target device has been opened
int MaestroController::controlTransfer(short serial, unsigned char RequestType, unsigned char Request, short Value,
                                        short Index)
{
  int ret = 0;
  switch (serial)
  {
    case (MICRO_SERIAL):
      ret = libusb_control_transfer(primary_hand, RequestType, Request, Value, Index, (unsigned char*)0, 0,
                                    (short)5000);
      break;
    case (PRIMARY_SERIAL):
      ret = libusb_control_transfer(primary_hand, RequestType, Request, Value, Index, (unsigned char*)0, 0,
                                    (short)5000);
      break;
    case (SECONDARY_SERIAL):
      ret = libusb_control_transfer(secondary_hand, RequestType, Request, Value, Index, (unsigned char*)0, 0,
                                    (short)5000);
      break;
  }
  if (ret)
  {
    ROS_ERROR("Control transfer failed.");
  }
  return ret;
}

// used for more complex communications with the controller, usually involving obtaining information from it
// mostly used for obtaining variable information, such as inputs and current servo states
// pre conditions: target device has been opened
int MaestroController::controlTransfer(short serial, unsigned char RequestType, unsigned char Request, short Value,
                                        short Index, void * data, short length)
{
  int ret = 0;

  switch (serial)
  {
    case (MICRO_SERIAL):
      ret = libusb_control_transfer(primary_hand, RequestType, Request, Value, Index, (unsigned char*)data, length,
                                    (short)5000);
      break;
    case (PRIMARY_SERIAL):
      ret = libusb_control_transfer(primary_hand, RequestType, Request, Value, Index, (unsigned char*)data, length,
                                    (short)5000);
      break;
    case (SECONDARY_SERIAL):
      ret = libusb_control_transfer(secondary_hand, RequestType, Request, Value, Index, (unsigned char*)data, length,
                                    (short)5000);
      break;
  }

  //TODO: remove or fix to fit all applications
  if (ret == (length - 24))
  {
    ROS_ERROR("SERVOS GET!");
  }
  return ret;
}
