/**
 * \file pelican_simulator.h
 *
 * \author Brian J. Julian
 *
 * \version 0.1
 *
 * \date 30 Aug 2011
 *
 */


#ifndef __PELICAN_SIMULATOR_H__
#define __PELICAN_SIMULATOR_H__


// includes
#include <ros/ros.h>
#include <boost/thread/thread.hpp>
#include <geometry_msgs/Vector3.h>
#include <geometry_msgs/Quaternion.h>
#include <wpi_msgs/LandRequest.h>
#include <wpi_msgs/LandConfirm.h>
#include <wpi_msgs/RobotState.h>


// define pi
#ifndef PI
#define PI 3.14159265358979323846
#endif

// define epsilon
#define EPSILON 1E-6

// macros
#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif

#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

#ifndef MINMAX
#define MINMAX(A,B,C) MIN((B),MAX((A),(C)))
#endif

#ifndef NORMALIZE
#define NORMALIZE(Z) atan2(sin(Z),cos(Z))
#endif


// enums
enum PelicanState
  {
    CHARGING,
    LAUNCHING,
    FLYING,
    FLYING_WITH_LOW_BATTERY,
    LANDING
  };


// namespaces
using namespace std;


// class
class PelicanSimulator
{
public:
 
  // constructor
  PelicanSimulator();

  // destructor
  ~PelicanSimulator();

private:

  // callbacks
  void callbackLandRequestMsg(const boost::shared_ptr<wpi_msgs::LandRequest const> &land_request_msg);
  void callbackLandConfirmMsg(const boost::shared_ptr<wpi_msgs::LandConfirm const> &land_confirm_msg);

   // init functions
  void initParams(void);
  void initPublishers(void);
  void initSubscribers(void);

  // robot states
  void chargingState(void);
  void launchingState(void);
  void flyingState(void);
  void flyingWithLowBatteryState(void);
  void landingState(void);

  // pelican thread
  void startPelicanThread(void);
  void pelicanThread(void);
  void stopPelicanThread(void);

   // robot state pub thread
  void startRobotStatePubThread(void);
  void robotStatePubThread(void);
  void stopRobotStatePubThread(void);

  // ros objects
  ros::NodeHandle nh_;
  ros::NodeHandle private_nh_;

  // ros param
  string uid_;
  double robot_state_pub_rate_;
  double pelican_thread_rate_;
  double flight_height_;
  double flight_speed_;
  double max_battery_voltage_;
  double return_battery_voltage_;
  double min_battery_voltage_;
  double battery_charge_rate_;
  double battery_discharge_rate_;
  
  // ros publishers
  ros::Publisher robot_state_pub_;
  ros::Publisher land_request_pub_;
  ros::Publisher land_confirm_pub_;

  // ros subscribers
  ros::Subscriber land_request_sub_;
  ros::Subscriber land_confirm_sub_;

  // threads
  boost::shared_ptr<boost::thread> robot_state_pub_thread_;
  boost::shared_ptr<boost::thread> pelican_thread_;
  boost::mutex mutex_;

  // general containers
  wpi_msgs::RobotState robot_state_;
  PelicanState pelican_state_;

  // volatile bools
  volatile bool received_land_request_;
  volatile bool received_positive_land_confirm_;
  volatile bool received_negative_land_confirm_;
};


#endif // __PELICAN_SIMULATOR_H__
