/**
 * \file pelican_simulator_state.cpp
 *
 * \author Brian J. Julian
 *
 * \version 0.1
 *
 * \date 30 Aug 2011
 *
 */


// includes
#include <pelican_simulator.h>


// charging state
void PelicanSimulator::chargingState(void)
{
  // motion
  // none

  // battery
  robot_state_.battery_voltage += battery_charge_rate_/pelican_thread_rate_;
  robot_state_.battery_voltage = MIN(robot_state_.battery_voltage, max_battery_voltage_);

  // state transitions
  if(robot_state_.battery_voltage > (max_battery_voltage_ - EPSILON))
    {
      // reset volatile bools
      received_land_request_ = false;
      received_positive_land_confirm_ = false;
      received_negative_land_confirm_ = false;

      // change state
      pelican_state_ = LAUNCHING;
    }
}


// launching state
void PelicanSimulator::launchingState(void)
{
  // motion
  robot_state_.translation.z += flight_speed_/pelican_thread_rate_;
  robot_state_.translation.z = MIN(robot_state_.translation.z, flight_height_);

  // battery
  robot_state_.battery_voltage -= battery_discharge_rate_/pelican_thread_rate_;
  robot_state_.battery_voltage = MAX(robot_state_.battery_voltage, min_battery_voltage_);

  // state transition
  if(robot_state_.translation.z > (flight_height_ - EPSILON))
    {
      // change state
      pelican_state_ = FLYING;
    }
  else if(received_land_request_)
    {
      // reset volatile bool
      received_land_request_ = false;

      // publish land confirm msg
      wpi_msgs::LandConfirm land_confirm_msg;
      land_confirm_msg.header.stamp = ros::Time::now();
      land_confirm_msg.uid = uid_;
      land_confirm_msg.confirm = true;
      land_confirm_pub_.publish(land_confirm_msg);

      // change state
      pelican_state_ = LANDING;
    }

}


// flying state
void PelicanSimulator::flyingState(void)
{
  // motion
  // none

  // battery
  robot_state_.battery_voltage -= battery_discharge_rate_/pelican_thread_rate_;
  robot_state_.battery_voltage = MAX(robot_state_.battery_voltage, min_battery_voltage_);

  // state transition
  if(received_land_request_)
    {
      // reset volatile bool
      received_land_request_ = false;

      // publish land confirm msg
      wpi_msgs::LandConfirm land_confirm_msg;
      land_confirm_msg.header.stamp = ros::Time::now();
      land_confirm_msg.uid = uid_;
      land_confirm_msg.confirm = true;
      land_confirm_pub_.publish(land_confirm_msg);

      // change state
      pelican_state_ = LANDING;
    }
  else if(robot_state_.battery_voltage < return_battery_voltage_)
    {
      // reset volatile bools
      received_positive_land_confirm_ = false;
      received_negative_land_confirm_ = false;

      // publish land request msg
      wpi_msgs::LandRequest land_request_msg;
      land_request_msg.header.stamp = ros::Time::now();
      land_request_msg.uid = uid_;
      land_request_pub_.publish(land_request_msg);

      // change state
      pelican_state_ = FLYING_WITH_LOW_BATTERY;
    }
}


// flying with low battery state
void PelicanSimulator::flyingWithLowBatteryState(void)
{
  // motion
  // none

  // battery
  robot_state_.battery_voltage -= battery_discharge_rate_/pelican_thread_rate_;
  robot_state_.battery_voltage = MAX(robot_state_.battery_voltage, min_battery_voltage_);

  // state transition
  if(received_land_request_)
    {
      // reset volatile bool
      received_land_request_ = false;

      // publish land confirm msg
      wpi_msgs::LandConfirm land_confirm_msg;
      land_confirm_msg.header.stamp = ros::Time::now();
      land_confirm_msg.uid = uid_;
      land_confirm_msg.confirm = true;
      land_confirm_pub_.publish(land_confirm_msg);

      // change state
      pelican_state_ = LANDING;
    }
  else if(received_positive_land_confirm_)
    {
      // reset volatile bool
      received_positive_land_confirm_ = false;

      // change state
      pelican_state_ = LANDING;
    }
  else if(received_negative_land_confirm_)
    {
      // reset volatile bool
      received_negative_land_confirm_ = false;

      // publish land request msg
      wpi_msgs::LandRequest land_request_msg;
      land_request_msg.header.stamp = ros::Time::now();
      land_request_msg.uid = uid_;
      land_request_pub_.publish(land_request_msg);
    }
}


// landing state
void PelicanSimulator::landingState(void)
{
  // motion
  robot_state_.translation.z -= flight_speed_/pelican_thread_rate_;
  robot_state_.translation.z = MAX(robot_state_.translation.z, 0.0);

  // battery
  robot_state_.battery_voltage -= battery_discharge_rate_/pelican_thread_rate_;
  robot_state_.battery_voltage = MAX(robot_state_.battery_voltage, min_battery_voltage_);

  // state transition
  if(robot_state_.translation.z < EPSILON)
    {
      // change state
      pelican_state_ = CHARGING;
    }
}
